SpeedyActions = LibStub("AceAddon-3.0"):NewAddon("SpeedyActions", "AceEvent-3.0")
SpeedyActions.supportModules = {}

local L = SpeedyActionsLocals
local _G = getfenv(0)
local CREATED_BUTTONS = 0
local overrideFrame = CreateFrame("Frame")
local buttonCache, cachedBindingButton, scannedKeys, hasEnteredWorld, foundNewKey = {}, {}, {}

function SpeedyActions:OnInitialize()
	self.defaults = {
		profile = {
			
		}	
	}
	
	self.db = LibStub:GetLibrary("AceDB-3.0"):New("SpeedyActionsDB", self.defaults)
	
	for name, module in pairs(self.supportModules) do
		if( IsAddOnLoaded(name) ) then
			module.loaded = true
			module:SupportLoaded()
		else
			self:RegisterEvent("ADDON_LOADED")
		end	
	end

	self:RegisterEvent("UPDATE_BINDINGS")
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:UPDATE_BINDINGS()
end	

-- Retrieves a button out of the cache
local function retrieveButton(attributeKey, attributeValue)
	local button
	for _, cacheButton in pairs(buttonCache) do
		if( not cacheButton.active ) then
			button = cacheButton
			break
		end
	end
	
	if( not button ) then
		CREATED_BUTTONS = CREATED_BUTTONS + 1
		button = CreateFrame("Button", "SpeedyActionsButtons" .. CREATED_BUTTONS, nil, "SecureActionButtonTemplate")
		table.insert(buttonCache, button)
	end
	
	button.active = true
	button.setKey = attributeKey
	button:SetAttribute(attributeKey, attributeValue)
	
	return button
end

-- Releases all buttons to cache
local function releaseButtons()
	for _, button in pairs(buttonCache) do
		if( button.active ) then
			button.active = nil
			
			if( button.setKey ) then
				button:SetAttribute(button.setKey, nil)
				button.setKey = nil
			end
		end	
	end
end

-- Rebinds the actual button
function SpeedyActions:RebindButton(buttonName, key, mouseButton)
	local button = _G[buttonName]
	if( button ) then
		button:RegisterForClicks("AnyDown")
		SetOverrideBindingClick(overrideFrame, true, key, buttonName, mouseButton)
	end
end

-- Takes the action bound to a key and figures out how it's going to have to be bound
function SpeedyActions:OverrideKeybind(key, action)
	if( action == "" ) then return end	
	
	-- SetBindingClick
	local buttonName, mouseButton = string.match(action, "^CLICK (.+):(.+)")
	if( buttonName ) then
		self:RebindButton(buttonName, key, mouseButton)
		return
	end
	
	-- SetBindingSpell
	local spell = string.match(action, "^SPELL (.+)")
	if( spell ) then
		self:RebindButton(retrieveButton("spell", spell), key)
		return
	end
	
	-- SetBindingItem
	local item = string.match(action, "^ITEM (.+)")
	if( item ) then
		self:RebindButton(retrieveButton("item", item), key)
		return
	end
	
	-- SetBindingMacro
	local macro = string.match(action, "^MACRO (.+)")
	if( macro ) then
		self:RebindButton(retrieveButton("macro", macro), key)
		return
	end
	
	-- SetBinding, but it's a multicast so it needs special handling
	local multiID = string.match(action, "MULTICASTSUMMONBUTTON(%d+)")
	if( multiID ) then
		self:RebindButton(retrieveButton("spell", TOTEM_MULTI_CAST_SUMMON_SPELLS[tonumber(multiID)]), key)
		return
	end
	
	-- SetBinding, for a totem recall
	local recallID = string.match(action, "MULTICASTRECALLBUTTON(%d+)")
	if( recallID ) then
		self:RebindButton(retrieveButton("spell", TOTEM_MULTI_CAST_RECALL_SPELLS[tonumber(recallID)]), key)
		return
	end
	
	-- None of those, it should be a default Blizzard one then, because Blizzard does not use the same casing
	-- as the frames name, the mass gsub to turn it into the buttons real name is necessary
	-- it's cached so it doesn't have to do this every time at least 
	local buttonName = cachedBindingButton[action]
	if( not buttonName ) then
		buttonName = string.gsub(action, "MULTIACTIONBAR4BUTTON", "MultiBarLeftButton")
		buttonName = string.gsub(buttonName, "MULTIACTIONBAR3BUTTON", "MultiBarRightButton")
		buttonName = string.gsub(buttonName, "MULTIACTIONBAR2BUTTON", "MultiBarBottomRightButton")
		buttonName = string.gsub(buttonName, "MULTIACTIONBAR1BUTTON", "MultiBarBottomLeftButton")
		buttonName = string.gsub(buttonName, "MULTICASTACTIONBUTTON", "MultiCastActionButton")
		buttonName = string.gsub(buttonName, "SHAPESHIFTBUTTON", "ShapeshiftButton")
		buttonName = string.gsub(buttonName, "PETACTIONBUTTON", "PetActionButon")
		buttonName = string.gsub(buttonName, "BONUSACTIONBUTTON", "BonusActionButton")
		buttonName = string.gsub(buttonName, "ACTIONBUTTON", "ActionButton")

		cachedBindingButton[action] = buttonName
	end
	
	self:RebindButton(buttonName, key)
end

-- One of the support modules found a key, register it then override whatever is associated with it (if we can)
function SpeedyActions:RegisterOverrideKey(...)
	for i=1, select("#", ...) do
		local key = select(i, ...)
		
		if( not scannedKeys[key] ) then foundNewKey = true end
		scannedKeys[key] = true
	end
end

function SpeedyActions:UPDATE_BINDINGS()
	foundNewKey = nil
	
	-- Can't update in combat, queue if so
	if( InCombatLockdown() ) then
		self:RegisterEvent("PLAYER_REGEN_ENABLED")
		return
	end
	
	ClearOverrideBindings(overrideFrame)
	releaseButtons()
	
	-- This adds support for bindings set to Blizzards default stuff, anything else is not supported
	-- and requires a support/ module
	for i=1, GetNumBindings() do
		local action, bindingOne, bindingTwo = GetBinding(i)
		if( bindingOne ) then scannedKeys[bindingOne] = true end
		if( bindingTwo ) then scannedKeys[bindingTwo] = true end
	end
	
	-- Tell the modules that there was a binding update and they need to look for any new supported keys
	for _, module in pairs(self.supportModules) do
		if( module.loaded ) then
			module:BindingsUpdated()
		end
	end
	
	-- Scan through all keys that were found using an action and override them
	for key in pairs(scannedKeys) do
		self:OverrideKeybind(key, GetBindingByKey(key))
	end
end

function SpeedyActions:PLAYER_ENTERING_WORLD()
	self:UnregisterEvent("PLAYER_ENTERING_WORLD")
	
	-- This means that between the initial bindings run and PLAYER_ENTERING_WORLD a support addon loaded
	-- that registered a key we didn't have yet, so everything needs to be rebound
	-- The only reason I do this is I'm not sure if PLAYER_REGEN_DISABLED can fire before PLAYER_ENTERING_WORLD
	-- need to verify this later
	if( foundNewKey ) then
		self:UPDATE_BINDINGS()
	end
	
	hasEnteredWorld = true
end

function SpeedyActions:ADDON_LOADED(event, addon)
	if( self.supportModules[addon] ) then
		self.supportModules[addon].loaded = true
		self.supportModules[addon]:SupportLoaded()
		
		-- A new key was found and the player has already entered the world, can't expect another binding update to occur
		if( foundNewKey and hasEnteredWorld ) then
			self:UPDATE_BINDINGS()
		end
	end
end

function SpeedyActions:PLAYER_REGEN_ENABLED()
	self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	self:UPDATE_BINDINGS()
end

--[===[@debug@
SpeedyActionsLocals = setmetatable(SpeedyActionsLocals, {
	__index = function(tbl, value)
		rawset(tbl, value, value)
		return value
	end,
})
--@end-debug@]===]